#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vi:ts=4:et

from . import localhost
import pycurl
import unittest
import gc
import weakref
try:
    import json
except ImportError:
    import simplejson as json

from . import appmanager
from . import util

setup_module, teardown_module = appmanager.setup(('app', 8380))

class DuphandleTest(unittest.TestCase):
    def setUp(self):
        self.orig = util.DefaultCurl()

    def test_duphandle_attribute_dict(self):
        self.orig.orig_attr = 'orig-value'
        # attribute dict should be copied - the *object*, not the reference
        dup = self.orig.duphandle()
        assert dup.orig_attr == 'orig-value'
        # cloned dict should be a separate object
        dup.dup_attr = 'dup-value'
        try:
            self.orig.dup_attr == 'does not exist'
        except AttributeError as error:
            assert 'trying to obtain a non-existing attribute: dup_attr' in str(error.args)
        else:
            self.fail('should have raised AttributeError')
        # dealloc self.orig - original dict is freed from memory
        self.orig.close()
        del self.orig
        # cloned dict should still exist
        assert dup.orig_attr == 'orig-value'
        assert dup.dup_attr == 'dup-value'
        dup.close()

    def slist_check(self, handle, value, persistance=True):
        body = util.BytesIO()
        handle.setopt(pycurl.WRITEFUNCTION, body.write)
        handle.setopt(pycurl.URL, 'http://%s:8380/header_utf8?h=x-test-header' % localhost)
        handle.perform()
        result = body.getvalue().decode('utf-8')
        assert (result == value) == persistance

    def slist_test(self, clear_func, *args):
        # new slist object is created with ref count = 1
        self.orig.setopt(pycurl.HTTPHEADER, ['x-test-header: orig-slist'])
        # ref is copied and object incref'ed
        dup1 = self.orig.duphandle()
        # slist object is decref'ed and ref set to null
        clear_func(*args)
        # null ref is copied - no effect
        dup2 = self.orig.duphandle()
        # check slist object persistance
        self.slist_check(dup1, 'orig-slist', True)
        self.slist_check(dup2, 'orig-slist', False)
        # check overwriting - orig slist is decref'ed to 0 and finally deallocated
        # util_curlslist_update() and util_curlslist_dealloc()
        dup1.setopt(pycurl.HTTPHEADER, ['x-test-header: dup-slist'])
        self.slist_check(dup1, 'dup-slist', True)
        # cleanup
        dup1.close()
        dup2.close()
        self.orig.close()

    def test_duphandle_slist_xdecref(self):
        # util_curl_xdecref()
        self.slist_test(self.orig.reset)

    def test_duphandle_slist_unsetopt(self):
        # util_curl_unsetopt()
        self.slist_test(self.orig.unsetopt, pycurl.HTTPHEADER)

    def httppost_check(self, handle, value, persistance=True):
        body = util.BytesIO()
        handle.setopt(pycurl.WRITEFUNCTION, body.write)
        handle.setopt(pycurl.URL, 'http://%s:8380/postfields' % localhost)
        handle.perform()
        result = json.loads(body.getvalue())
        assert (result == value) == persistance

    def httppost_test(self, clear_func, *args):
        self.orig.setopt(pycurl.HTTPPOST, [
            ('field', (pycurl.FORM_CONTENTS, 'orig-httppost')),
        ])
        dup1 = self.orig.duphandle()
        clear_func(*args)
        dup2 = self.orig.duphandle()
        self.httppost_check(dup1, {'field': 'orig-httppost'}, True)
        self.httppost_check(dup2, {'field': 'orig-httppost'}, False)
        # util_curlhttppost_update() and util_curlhttppost_dealloc()
        dup1.setopt(pycurl.HTTPPOST, [
            ('field', (pycurl.FORM_CONTENTS, 'dup-httppost')),
        ])
        self.httppost_check(dup1, {'field': 'dup-httppost'}, True)
        dup1.close()
        dup2.close()
        self.orig.close()

    def test_duphandle_httppost_xdecref(self):
        # util_curl_xdecref()
        self.httppost_test(self.orig.reset)

    def test_duphandle_httppost_unsetopt(self):
        # util_curl_unsetopt()
        self.httppost_test(self.orig.unsetopt, pycurl.HTTPPOST)

    def test_duphandle_references(self):
        body = util.BytesIO()
        def callback(data):
            body.write(data)
        callback_ref = weakref.ref(callback)
        # preliminary checks of gc and weakref working as expected
        assert gc.get_referrers(callback) == []
        assert callback_ref() is not None
        # setopt - callback ref is copied and callback incref'ed
        self.orig.setopt(pycurl.WRITEFUNCTION, callback)
        assert gc.get_referrers(callback) == [self.orig]
        # duphandle - callback ref is copied and callback incref'ed
        dup = self.orig.duphandle()
        assert set(gc.get_referrers(callback)) == {self.orig, dup}
        # dealloc self.orig and decref callback
        self.orig.close()
        del self.orig
        assert gc.get_referrers(callback) == [dup]
        # decref callback again - back to ref count = 1
        del callback
        assert callback_ref() is not None
        # check that callback object still exists and is invoked
        dup.setopt(pycurl.URL, 'http://%s:8380/success' % localhost)
        dup.perform()
        result = body.getvalue().decode('utf-8')
        assert result == 'success'
        # final decref - callback is deallocated
        dup.close()
        assert callback_ref() is None



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vi:ts=4:et

from . import localhost
import pycurl
import pytest
import sys
import unittest

from . import appmanager
from . import util

setup_module, teardown_module = appmanager.setup(('app', 8380))

class MultiCallbackTest(unittest.TestCase):
    def setUp(self):
        self.easy = util.DefaultCurl()
        self.easy.setopt(pycurl.URL, 'http://%s:8380/long_pause' % localhost)
        self.multi = pycurl.CurlMulti()
        self.multi.setopt(pycurl.M_SOCKETFUNCTION, self.socket_callback)
        self.multi.setopt(pycurl.M_TIMERFUNCTION, self.timer_callback)
        self.socket_result = None
        self.timer_result = None
        self.sockets = {}

    def tearDown(self):
        self.multi.close()
        self.easy.close()

    def socket_callback(self, ev_bitmask, sock_fd, multi, data):
        self.socket_result = (sock_fd, ev_bitmask)
        if ev_bitmask & pycurl.POLL_REMOVE:
            self.sockets.pop(sock_fd)
        else:
            self.sockets[sock_fd] = ev_bitmask | self.sockets.get(sock_fd, 0)

    def timer_callback(self, timeout_ms):
        self.timer_result = timeout_ms

    def partial_transfer(self):
        perform = True
        def write_callback(data):
            nonlocal perform
            perform = False
        self.easy.setopt(pycurl.WRITEFUNCTION, write_callback)
        self.multi.add_handle(self.easy)
        self.multi.socket_action(pycurl.SOCKET_TIMEOUT, 0)
        while self.sockets and perform:
            for socket, action in tuple(self.sockets.items()):
                self.multi.socket_action(socket, action)

    # multi.socket_action must call both SOCKETFUNCTION and TIMERFUNCTION at
    # various points during the transfer (at least at the start and end)
    @pytest.mark.xfail(sys.platform == 'darwin', reason='https://github.com/pycurl/pycurl/issues/729')
    def test_multi_socket_action(self):
        self.multi.add_handle(self.easy)
        self.timer_result = None
        self.socket_result = None
        self.multi.socket_action(pycurl.SOCKET_TIMEOUT, 0)
        assert self.socket_result is not None
        assert self.timer_result is not None

    # multi.add_handle must call TIMERFUNCTION to schedule a kick-start
    def test_multi_add_handle(self):
        self.multi.add_handle(self.easy)
        assert self.timer_result is not None

    # (mid-transfer) multi.remove_handle must call SOCKETFUNCTION to remove sockets
    @pytest.mark.xfail(sys.platform == 'darwin', reason='https://github.com/pycurl/pycurl/issues/729')
    def test_multi_remove_handle(self):
        self.multi.add_handle(self.easy)
        self.multi.socket_action(pycurl.SOCKET_TIMEOUT, 0)
        self.socket_result = None
        self.multi.remove_handle(self.easy)
        assert self.socket_result is not None

    # (mid-transfer) easy.pause(PAUSE_ALL) must call SOCKETFUNCTION to remove sockets
    # (mid-transfer) easy.pause(PAUSE_CONT) must call TIMERFUNCTION to resume
    @pytest.mark.xfail(sys.platform == 'darwin', reason='https://github.com/pycurl/pycurl/issues/729')
    def test_easy_pause_unpause(self):
        self.partial_transfer()
        self.socket_result = None
        # libcurl will now inform us that we should remove some sockets
        self.easy.pause(pycurl.PAUSE_ALL)
        assert self.socket_result is not None
        self.socket_result = None
        self.timer_result = None
        # libcurl will now tell us to add those sockets and schedule a kickstart
        self.easy.pause(pycurl.PAUSE_CONT)
        assert self.socket_result is not None
        assert self.timer_result is not None

    # (mid-transfer) easy.close() must call SOCKETFUNCTION to remove sockets
    @pytest.mark.xfail(sys.platform == 'darwin', reason='https://github.com/pycurl/pycurl/issues/729')
    def test_easy_close(self):
        self.partial_transfer()
        self.socket_result = None
        self.easy.close()
        assert self.socket_result is not None
